# -*- coding: utf-8 -*-
u"""
:mod:`edx_ace.monitoring` exposes functions that are useful for reporting ACE
message delivery stats to monitoring services.
"""
from __future__ import absolute_import

try:
    import newrelic.agent
except ImportError:
    newrelic = None  # pylint: disable=invalid-name


def report(key, value):
    report_to_newrelic(key, value)


def report_to_newrelic(key, value):
    if newrelic:
        newrelic.agent.add_custom_parameter(key, value)



# -*- coding: utf-8 -*-
u"""
:mod:`edx_ace.utils.plugins` contains utility functions used
to make working with the ACE plugin system easier. These are intended
for internal use by ACE.
"""
from __future__ import absolute_import, division, print_function

import logging
from functools import partial

from stevedore import enabled

LOG = logging.getLogger(__name__)


def get_manager(namespace, names=None):
    u"""
    Get the stevedore extension manager for this namespace.

    Args:
        namespace (basestring): The entry point namespace to load plugins for.
        names (list): A list of names to load. If this is ``None`` then all extension will be loaded from this
            namespace.

    Returns:
        stevedore.enabled.EnabledExtensionManager: Extension manager with all extensions instantiated.
    """
    return enabled.EnabledExtensionManager(
        namespace=namespace,
        check_func=partial(check_plugin, namespace=namespace, names=names),
        invoke_on_load=True,
    )


def get_plugins(namespace, names=None):
    u"""
    Get all extensions for this namespace and list of names.

    Args:
        namespace (basestring): The entry point namespace to load plugins for.
        names (list): A list of names to load. If this is ``None`` then all extension will be loaded from this
            namespace.

    Returns:
        list: A list of extensions.
    """
    return list(get_manager(namespace, names))


def check_plugin(extension, namespace, names=None):
    u"""
    Check the extension to see if it's enabled.

    Args:
        extension (stevedore.extension.Extension): The extension to check.
        namespace (basestring): The namespace that the extension was loaded from.
        names (list): A whitelist of extensions that should be checked.

    Returns:
        bool: Whether or not this extension is enabled and should be used.
    """
    if names is None or extension.name in names:
        plugin_enabled = extension.plugin.enabled()
        if not plugin_enabled:
            LOG.info(u'Extension with name %s for namespace %s is not enabled', extension.name, namespace)
        return plugin_enabled
    return False



u"""
Tests of :mod:`edx_ace.utils.date`.
"""
from __future__ import absolute_import

from datetime import datetime
from unittest import TestCase

from hypothesis import example, given
from hypothesis import strategies as st
from hypothesis.extra.pytz import timezones

from edx_ace.utils.date import deserialize, serialize


class TestDateSerialization(TestCase):
    """ Test Date Serialization. """
    @given(st.one_of(st.datetimes(timezones=st.none() | timezones()), st.none()))
    @example(datetime(16, 1, 1, 0, 0, 16))
    def test_round_trip(self, date):
        serialized = serialize(date)
        parsed = deserialize(serialized)
        self.assertEqual(date, parsed)



# -*- coding: utf-8 -*-
u"""
:mod:`edx_ace.message` contains the core :class:`Message` and :class:`MessageType`
classes, which allow specification of the content to be delivered by ACE.
"""
from __future__ import absolute_import, division, print_function

import logging
from abc import ABCMeta
from uuid import UUID, uuid4

import attr
import six

from django.apps import apps

import edx_ace.utils.date as date
from edx_ace.monitoring import report as monitoring_report
from edx_ace.recipient import Recipient
from edx_ace.serialization import MessageAttributeSerializationMixin


@attr.s
@six.add_metaclass(ABCMeta)
class Message(MessageAttributeSerializationMixin):
    u"""
    A ``Message`` is the core piece of data that is passed into ACE.
    It captures the message, recipient, and all context needed to render
    the message for delivery.

    Arguments:
        app_label (str): The name of the Django app that is sending
            this message. Used to look up the appropriate template
            during rendering. Required.
        name (str): The name of this type of message. Used to look up
            the appropriate template during rendering. Required.

        recipient (:class:`.Recipient`): The intended recipient of the
            message. Optional.
        expiration_time(:class:`~datetime.datetime`): The date and time
            at which this message expires. After this time, the message
            should not be delivered. Optional.
        context (dict): A dictionary to be supplied to the template at
            render time as the context.
        send_uuid (:class:`uuid.UUID`): The :class:`uuid.UUID` assigned
            to this bulk-send of many messages.
        language (str): The language the message should be rendered in.
            Optional.
    """

    # mandatory attributes
    # Name is the unique identifier for the message type.
    # Used for:
    #   tracking
    #   discovering message presentation templates
    app_label = attr.ib()
    name = attr.ib()

    # optional attributes
    recipient = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(Recipient)),
    )

    expiration_time = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(date.datetime)),
    )

    context = attr.ib()

    # TODO(later): better naming to distinguish between these 2 UUIDs
    uuid = attr.ib(
        init=False,
        validator=attr.validators.instance_of(UUID),
    )
    send_uuid = attr.ib(
        validator=attr.validators.optional(attr.validators.instance_of(UUID)),
        default=None
    )
    options = attr.ib()
    language = attr.ib(default=None)
    log_level = attr.ib(default=None)

    @context.default
    def default_context_value(self):
        return {}

    @options.default
    def default_options_value(self):
        return {}

    @uuid.default
    def generate_uuid(self):
        return uuid4()

    @property
    def unique_name(self):
        u"""
        A unique name for this message, used for logging and reporting.

        Returns: str
        """
        return u'.'.join([self.app_label, self.name])

    @property
    def log_id(self):
        u"""
        The identity of this message for logging.
        """
        return u'.'.join([
            self.unique_name,
            six.text_type(self.send_uuid) if self.send_uuid else u'no_send_uuid',
            six.text_type(self.uuid)
        ])

    def get_message_specific_logger(self, logger):
        u"""
        Arguments:
            logger (:class:`logging.Logger`): The logger to be adapted.

        Returns:  :class:`.MessageLoggingAdapter` that is specific to this message.
        """
        return MessageLoggingAdapter(logger, {u'message': self})

    def report_basics(self):
        monitoring_report(u'message_name', self.unique_name)
        monitoring_report(u'language', self.language)

    def report(self, key, value):
        monitoring_report(key, value)


class MessageLoggingAdapter(logging.LoggerAdapter):
    u"""
    A :class:`logging.LoggingAdapter` that prefixes log items with
    a message :attr:`.log_id`.ABCMeta

    Expects a ``message`` key in its ``extra`` argument which should
    contain the :class:`.Message` being logged for.
    """
    def process(self, msg, kwargs):
        return u'[%s] %s' % (self.extra[u'message'].log_id, msg), kwargs

    def debug(self, msg, *args, **kwargs):
        log_level = self.extra[u'message'].log_level
        if log_level and log_level <= logging.DEBUG:
            self.info(msg, *args, **kwargs)
        else:
            super(MessageLoggingAdapter, self).debug(msg, *args, **kwargs)


@attr.s(eq=False, order=False)
class MessageType(MessageAttributeSerializationMixin):
    u"""
    A class representing a type of :class:`Message`. An instance of
    a ``MessageType`` is used for each batch send of messages.

    Arguments:
        context (dict): Context to be supplied to all messages sent in
            this batch of messages.
        expiration_time (:class:`datetime.datetime`): The time at which
            these messages expire.
        app_label (str): Override the Django app that is used to resolve
            the template for rendering. Defaults to :attr:`.APP_LABEL` or
            to the app that the message type was defined in.
        name (str): Override the message name that is used to resolve
            the template for rendering. Defaults to :attr:`.NAME` or
            to the name of the class.
    """
    NAME = None
    APP_LABEL = None

    context = attr.ib()
    uuid = attr.ib(
        init=False,
        validator=attr.validators.instance_of(UUID),
    )
    expiration_time = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(date.datetime)),
    )
    app_label = attr.ib()
    name = attr.ib()
    options = attr.ib()
    log_level = attr.ib(default=None)

    @context.default
    def default_context_value(self):
        return {}

    @options.default
    def default_options_value(self):
        return {}  # pragma: no cover

    @uuid.default
    def generate_uuid(self):
        return uuid4()

    @name.default
    def default_name(self):
        """ Return default class name. """
        if self.NAME is None:
            return self.__class__.__name__.lower()
        else:
            return self.NAME

    @app_label.default
    def default_app_label(self):
        u""" Get default app Label. """
        if self.APP_LABEL is None:
            return apps.get_containing_app_config(self.__class__.__module__).label
        else:
            return self.APP_LABEL

    def personalize(self, recipient, language, user_context):
        u"""
        Personalize this `MessageType` to a specific recipient, in order to
        send a specific message.

        Arguments:
            recipient (:class:`.Recipient`): The intended recipient of the
                message. Optional.
            language (str): The language the message should be rendered in.
                Optional.
            user_context (dict): A dictionary containing recipient-specific
                context to be supplied to the template at render time.

        Returns: A new :class:`.Message` that has been personalized to a
            specific recipient.
        """
        context = dict(self.context)
        context.update(user_context)
        return Message(
            app_label=self.app_label,
            name=self.name,
            expiration_time=self.expiration_time,
            context=context,
            send_uuid=self.uuid,
            recipient=recipient,
            language=language,
            log_level=self.log_level,
            options=self.options,
        )

    # We override these so that a subtype of MessageType can compare equal
    # to a deserialized non-subtyped MessageType
    def __eq__(self, other):
        if isinstance(other, MessageType):
            return attr.astuple(self) == attr.astuple(other)
        else:
            return NotImplemented

    def __hash__(self):
        return hash(attr.astuple(self))

    def __ne__(self, other):
        result = self == other
        if result == NotImplemented:
            return NotImplemented
        else:
            return not result



u"""
ACE (Automated Communications Engine) is a framework for automatically
sending messages to users.

:mod:`edx_ace` exports the typical set of functions and classes needed to use
ACE.
"""
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from .ace import send
from .channel import Channel, ChannelType
from .message import Message, MessageType
from .policy import Policy, PolicyResult
from .recipient import Recipient
from .recipient_resolver import RecipientResolver

__version__ = u'0.1.14'

default_app_config = u'edx_ace.apps.EdxAceConfig'

__all__ = [
    u'send',
    u'__version__',
    u'Message',
    u'MessageType',
    u'Recipient',
    u'RecipientResolver',
    u'ChannelType',
    u'Channel',
    u'Policy',
    u'PolicyResult',
]



u"""
Tests of :mod:`edx_ace.presentation`.
"""
from __future__ import absolute_import

from unittest import TestCase

from mock import Mock

from edx_ace.channel import ChannelType
from edx_ace.errors import UnsupportedChannelError
from edx_ace.presentation import render


class TestRender(TestCase):
    u"""
    Tests for unsupported rendering for a channel.
    """

    def test_missing_renderer(self):
        channel = Mock(
            channel_type=ChannelType.PUSH,
        )

        message = Mock()

        with self.assertRaises(UnsupportedChannelError):
            render(channel, message)



u"""
Tests of :mod:`edx_ace.channel`.
"""
from __future__ import absolute_import

from mock import patch

from django.test import TestCase, override_settings

from edx_ace.channel import ChannelMap, ChannelType, get_channel_for_message
from edx_ace.channel.file import FileEmailChannel
from edx_ace.channel.sailthru import SailthruEmailChannel
from edx_ace.errors import UnsupportedChannelError
from edx_ace.message import Message
from edx_ace.recipient import Recipient
from edx_ace.utils.date import get_current_time


class TestChannelMap(TestCase):
    u"""
    Tests for the channels().
    """

    def setUp(self):
        self.msg_kwargs = {
            u'app_label': u'test_app_label',
            u'name': u'test_message',
            u'expiration_time': get_current_time(),
            u'context': {
                u'key1': u'value1',
                u'key2': u'value2',
            },
            u'recipient': Recipient(
                username=u'me',
            )
        }

    def test_get_channel_for_message(self):
        channel_map = ChannelMap([
            [u'file_email', FileEmailChannel],
            [u'sailthru_email', SailthruEmailChannel],
        ])

        transactional_msg = Message(options={u'transactional': True}, **self.msg_kwargs)
        info_msg = Message(options={}, **self.msg_kwargs)

        with patch(u'edx_ace.channel.channels', return_value=channel_map):
            assert get_channel_for_message(ChannelType.EMAIL, transactional_msg) is FileEmailChannel
            assert get_channel_for_message(ChannelType.EMAIL, info_msg) is SailthruEmailChannel

            with self.assertRaises(UnsupportedChannelError):
                assert get_channel_for_message(ChannelType.PUSH, transactional_msg)

    @override_settings(
        ACE_CHANNEL_DEFAULT_EMAIL='sailthru_email',
        ACE_CHANNEL_TRANSACTIONAL_EMAIL='file_email',
    )
    def test_default_channel(self):
        channel_map = ChannelMap([
            ['sailthru', SailthruEmailChannel],
        ])

        message = Message(options={u'transactional': True}, **self.msg_kwargs)

        with patch(u'edx_ace.channel.channels', return_value=channel_map):
            channel = get_channel_for_message(ChannelType.EMAIL, message)
            assert channel is SailthruEmailChannel



# -*- coding: utf-8 -*-
u"""
:mod:`edx_ace.utils.once` provides the ability to create a module-level
function that caches it's result after the first call (this can be used
for lazy-loading expensive computations).
"""
from __future__ import absolute_import, division, print_function

import functools


def once(func):
    u"""
    Decorates a function that will be called exactly once.

    After the function is called once, its result is stored in memory and immediately returned to subsequent callers
    instead of calling the decorated function again.

    Examples:

        An incrementing value::

            _counter = 0

            @once
            def get_counter():
                global _counter
                _counter += 1
                return _counter

            def get_counter_updating():
                global _counter
                _counter += 1
                return _counter

            print(get_counter())  # This will print "0"
            print(get_counter_updating()) # This will print "1"
            print(get_counter())  # This will also print "0"
            print(get_counter_updating()) # This will print "2"

        Lazy loading::

            @once
            def load_config():
                with open('config.json', 'r') as cfg_file:
                    return json.load(cfg_file)

            cfg = load_config()  # This will do the relatively expensive operation to
                                 # read the file from disk.
            cfg2 = load_config() # This call will not reload the file from disk, it
                                 # will use the value returned by the first invocation
                                 # of this function.

    Args:
        func (callable): The function that should be called exactly once.

    Returns:
        callable: The wrapped function.
    """

    @functools.wraps(func)
    def wrapper():
        if not hasattr(func, u'__once_result'):
            func.__once_result = func()  # pylint: disable=protected-access
        return func.__once_result  # pylint: disable=protected-access

    return wrapper



# -*- coding: utf-8 -*-
u"""
Internal module configuring ACE as a Django app.
"""
from __future__ import absolute_import, division, print_function

from django.apps import AppConfig


class EdxAceConfig(AppConfig):
    u"""
    Configuration for the edx_ace Django application.
    """

    name = u'edx_ace'



# -*- coding: utf-8 -*-
u"""
The main entry point for sending messages with ACE.

Usage:

.. code-block:: python

    from edx_ace import ace
    from edx_ace.messages import Message

    msg = Message(
        name="test_message",
        app_label="my_app",
        recipient=Recipient(username='a_user', email='a_user@example.com'),
        language='en',
        context={
            'stuff': 'to personalize the message',
        }
    )
    ace.send(msg)
"""

from __future__ import absolute_import, division, print_function

import six

from edx_ace import delivery, policy, presentation
from edx_ace.channel import get_channel_for_message
from edx_ace.errors import ChannelError, UnsupportedChannelError


def send(msg):
    u"""
    Send a message to a recipient.

    Calling this method will result in an attempt being made to deliver the provided message to the recipient. Depending
    on the configured policies, it may be transmitted to them over one or more channels (email, sms, push etc).

    The message must have valid values for all required fields in order for it to be sent. Different channels have
    different requirements, so care must be taken to ensure that all of the needed information is present in the message
    before calling ``ace.send()``.

    Args:
        msg (Message): The message to send.
    """
    msg.report_basics()

    channels_for_message = policy.channels_for(msg)

    for channel_type in channels_for_message:
        try:
            channel = get_channel_for_message(channel_type, msg)
        except UnsupportedChannelError:
            continue

        try:
            rendered_message = presentation.render(channel, msg)
            delivery.deliver(channel, rendered_message, msg)
        except ChannelError as error:
            msg.report(
                u'{channel_type}_error'.format(channel_type=channel_type),
                six.text_type(error)
            )



# pylint: disable=missing-docstring
from __future__ import absolute_import

import datetime

from dateutil.tz import tzutc
from mock import Mock, call, patch, sentinel

from django.test import TestCase

from edx_ace.channel import ChannelType
from edx_ace.delivery import deliver
from edx_ace.errors import FatalChannelDeliveryError, RecoverableChannelDeliveryError
from edx_ace.message import Message
from edx_ace.recipient import Recipient


class TestDelivery(TestCase):
    def setUp(self):
        super(TestDelivery, self).setUp()

        self.mock_channel = Mock(
            name=u'test_channel',
            channel_type=ChannelType.EMAIL
        )
        self.recipient = Recipient(
            username=str(sentinel.username)
        )
        self.message = Message(
            app_label=str(sentinel.app_label),
            name=str(sentinel.name),
            options={
                u'from_address': u'bulk@example.com',
            },
            recipient=self.recipient,
        )
        self.current_time = datetime.datetime.utcnow().replace(tzinfo=tzutc())

    def test_happy_path(self):
        deliver(self.mock_channel, sentinel.rendered_email, self.message)
        self.mock_channel.deliver.assert_called_once_with(self.message, sentinel.rendered_email)

    def test_fatal_error(self):
        self.mock_channel.deliver.side_effect = FatalChannelDeliveryError(u'testing')
        with self.assertRaises(FatalChannelDeliveryError):
            deliver(self.mock_channel, sentinel.rendered_email, self.message)

    @patch(u'edx_ace.delivery.get_current_time')
    def test_custom_message_expiration(self, mock_get_current_time):
        self.message.expiration_time = self.current_time - datetime.timedelta(seconds=10)
        mock_get_current_time.return_value = self.current_time
        deliver(self.mock_channel, sentinel.rendered_email, self.message)
        assert not self.mock_channel.deliver.called

    @patch(u'edx_ace.delivery.time')
    @patch(u'edx_ace.delivery.get_current_time')
    def test_single_retry(self, mock_get_current_time, mock_time):
        mock_get_current_time.side_effect = [
            self.current_time,  # First call to figure out the expiration time
            self.current_time,  # Check to see if the message has expired
            self.current_time,  # Time after attempting the send
            self.current_time + datetime.timedelta(seconds=1.1),
        ]
        self.mock_channel.deliver.side_effect = [
            RecoverableChannelDeliveryError(u'Try again later', self.current_time + datetime.timedelta(seconds=1)),
            True
        ]
        deliver(self.mock_channel, sentinel.rendered_email, self.message)
        assert self.mock_channel.deliver.call_count == 2
        mock_time.sleep.assert_called_once_with(1)

    @patch(u'edx_ace.delivery.time')
    @patch(u'edx_ace.delivery.get_current_time')
    def test_next_attempt_time_after_expiration(self, mock_get_current_time, mock_time):
        self.message.expiration_time = self.current_time + datetime.timedelta(seconds=10)
        mock_get_current_time.return_value = self.current_time
        self.mock_channel.deliver.side_effect = [
            RecoverableChannelDeliveryError(u'Try again later', self.current_time + datetime.timedelta(seconds=11)),
        ]
        deliver(self.mock_channel, sentinel.rendered_email, self.message)
        assert not mock_time.sleep.called
        assert self.mock_channel.deliver.call_count == 1

    @patch(u'edx_ace.delivery.time')
    @patch(u'edx_ace.delivery.get_current_time')
    def test_multiple_retries(self, mock_get_current_time, mock_time):
        mock_get_current_time.side_effect = [
            self.current_time,  # First call to figure out the expiration time
            self.current_time,  # Check to see if the message has expired
            self.current_time,  # Time after attempting the send
            # This attempt fails and the app is instructed to wait for a second
            self.current_time + datetime.timedelta(seconds=1),  # Check again to see if it's expired
            self.current_time + datetime.timedelta(seconds=1),  # Time after second send attempt
            # This attempt fails and the app is instructed to wait for another second
            self.current_time + datetime.timedelta(seconds=2),  # Final expiration check
        ]
        self.mock_channel.deliver.side_effect = [
            RecoverableChannelDeliveryError(u'Try again later', self.current_time + datetime.timedelta(seconds=1)),
            RecoverableChannelDeliveryError(u'Try again later', self.current_time + datetime.timedelta(seconds=2)),
            True
        ]
        deliver(self.mock_channel, sentinel.rendered_email, self.message)
        assert mock_time.sleep.call_args_list == [call(1), call(1)]
        assert self.mock_channel.deliver.call_count == 3



u"""
Tests of :mod:`edx_ace.ace`.
"""
from __future__ import absolute_import

from mock import Mock, patch

from django.test import TestCase

from edx_ace import ace
from edx_ace.channel import ChannelMap, ChannelType
from edx_ace.errors import UnsupportedChannelError
from edx_ace.message import Message
from edx_ace.recipient import Recipient
from edx_ace.renderers import RenderedEmail
from edx_ace.test_utils import StubPolicy, patch_policies


class TestAce(TestCase):
    u"""
    Tests for the send method.
    """
    def test_ace_send_happy_path(self):
        patch_policies(self, [StubPolicy([ChannelType.PUSH])])
        mock_channel = Mock(
            channel_type=ChannelType.EMAIL,
            action_links=[],
            tracker_image_sources=[],
        )

        recipient = Recipient(username=u'testuser')
        msg = Message(
            app_label=u'testapp',
            name=u'testmessage',
            recipient=recipient,
        )

        channel_map = ChannelMap([
            [u'sailthru_email', mock_channel],
        ])

        with patch(u'edx_ace.channel.channels', return_value=channel_map):
            ace.send(msg)

        mock_channel.deliver.assert_called_once_with(
            msg,
            RenderedEmail(
                from_name=u'template from_name.txt',
                subject=u'template subject.txt',
                # The new lines are needed because the template has some tags, which means leftover newlines
                body_html=u'template body.html\n\n\n\n\n',
                head_html=u'template head.html\n',
                body=u'template body.txt',
            ),
        )

    @patch(u'edx_ace.ace.get_channel_for_message', side_effect=UnsupportedChannelError)
    def test_ace_send_unsupported_channel(self, *_args):
        recipient = Recipient(username=u'testuser')
        msg = Message(
            app_label=u'testapp',
            name=u'testmessage',
            recipient=recipient,
        )

        ace.send(msg)  # UnsupportedChannelError shouldn't throw UnsupportedChannelError



# pylint: disable=missing-docstring
from __future__ import absolute_import

from smtplib import SMTPException

from mock import Mock, patch

from django.core import mail
from django.test import TestCase, override_settings

from edx_ace.channel.django_email import DjangoEmailChannel
from edx_ace.errors import FatalChannelDeliveryError
from edx_ace.message import Message
from edx_ace.presentation import render
from edx_ace.recipient import Recipient


class TestDjangoEmailChannel(TestCase):
    def setUp(self):
        super(TestDjangoEmailChannel, self).setUp()

        self.channel = DjangoEmailChannel()
        self.message = Message(
            app_label=u'testapp',
            name=u'testmessage',
            options={
                u'from_address': u'bulk@example.com',
            },
            recipient=Recipient(username=u'Robot', email_address=u'mr@robot.io'),
        )

        self.mock_rendered_message = Mock(
            subject=u'\n Hello from  \r\nRobot ! \n',
            body=u'Just trying to see what is like to talk to a human!',
            body_html=u"""
                <p>Just trying to see what is like to talk to a human!</p>

                <hr />
            """,
        )

    def _get_rendered_message(self):
        channel = DjangoEmailChannel()
        message = Message(
            app_label=u'testapp',
            name=u'testmessage',
            options={},
            recipient=Recipient(username=u'Robot', email_address=u'mr@robot.io'),
        )

        return render(channel, message)

    def test_enabled_method(self):
        assert self.channel.enabled()

    def test_happy_path(self):
        self.channel.deliver(self.message, self.mock_rendered_message)

        sent_email = mail.outbox[0]

        html_body, _ = sent_email.alternatives[0]

        assert sent_email.subject == u'Hello from Robot !'
        assert u'Just trying to see what' in sent_email.body
        assert sent_email.to == [u'mr@robot.io']

        assert u'talk to a human!</p>' in html_body

    @patch(u'django.core.mail.EmailMultiAlternatives.send', side_effect=SMTPException)
    def test_smtp_failure(self, _send):
        with self.assertRaises(FatalChannelDeliveryError):
            self.channel.deliver(self.message, self.mock_rendered_message)

    @override_settings(DEFAULT_FROM_EMAIL=None)
    def test_with_no_from_address_without_default(self):
        message = Message(
            app_label=u'testapp',
            name=u'testmessage',
            options={},
            recipient=Recipient(username=u'Robot', email_address=u'mr@robot.io'),
        )

        with self.assertRaises(FatalChannelDeliveryError):
            self.channel.deliver(message, self.mock_rendered_message)

    @override_settings(DEFAULT_FROM_EMAIL=u'hello@edx.org')
    def test_with_no_from_address_with_default(self):
        message = Message(
            app_label=u'testapp',
            name=u'testmessage',
            options={},
            recipient=Recipient(username=u'Robot', email_address=u'mr@robot.io'),
        )

        self.channel.deliver(message, self.mock_rendered_message)
        assert len(mail.outbox) == 1, u'Should have one email'

    def test_render_email_with_django_channel(self):
        rendered_email = self._get_rendered_message()
        assert u'{beacon_src}' not in rendered_email.body_html
        assert u'{view_url}' not in rendered_email.body_html
        assert u'{optout_confirm_url}' not in rendered_email.body_html

    def test_happy_sending_rendered_email(self):
        rendered_message = self._get_rendered_message()
        self.channel.deliver(self.message, rendered_message)

        assert mail.outbox, u'Should send the message'

        html_email = mail.outbox[0].alternatives[0][0]
        assert u'template head.html' in html_email
        assert u'template body.html' in html_email

    def test_happy_email_with_reply_to(self):
        rendered_message = self._get_rendered_message()
        self.message.options[u'reply_to'] = [u'learner@example.com']
        self.channel.deliver(self.message, rendered_message)

        sent_email = mail.outbox[0]

        assert sent_email.reply_to == [u'learner@example.com']



# pylint: disable=missing-docstring
from __future__ import absolute_import

import ddt
from mock import patch

from django.test import TestCase, override_settings

from edx_ace.channel.sailthru import SailthruEmailChannel
from edx_ace.delivery import deliver
from edx_ace.message import Message
from edx_ace.presentation import render
from edx_ace.recipient import Recipient


@ddt.ddt
class TestSailthruChannel(TestCase):

    def setUp(self):
        self.channel = SailthruEmailChannel()

    def test_render_email_with_sailthru(self):
        message = Message(
            app_label=u'testapp',
            name=u'testmessage',
            options={},
            recipient=Recipient(username=u'Robot', email_address=u'mr@robot.io'),
        )

        rendered_email = render(self.channel, message)

        assert u'{beacon_src}' in rendered_email.body_html
        assert u'{view_url}' in rendered_email.body_html
        assert u'{optout_confirm_url}' in rendered_email.body_html

    @override_settings(
        ACE_CHANNEL_SAILTHRU_DEBUG=False,
    )
    @ddt.data(
        ({u'from_address': 'custom@example.com'}, {u'behalf_email': 'custom@example.com'}),
        ({u'reply_to': 'student@example.com'}, {u'behalf_email': 'student@example.com'}),
        ({u'irrelevant': 'test'}, {}),
        ({}, {}),
    )
    @ddt.unpack
    def test_on_behalf_option_with_sailthru(self, message_options, expected_options):
        """
        Tests sailthru send API is called with on_behalf option
        """
        message = Message(
            app_label=u'testapp',
            name=u'testmessage',
            options=message_options,
            recipient=Recipient(username=u'Robot', email_address=u'mr@robot.io'),
        )
        rendered_email = render(self.channel, message)

        with patch('edx_ace.channel.sailthru.SailthruClient.send') as mock_send:
            deliver(self.channel, rendered_email, message)
            self.assertEqual(mock_send.call_args_list[0][1]['options'], expected_options)



# -*- coding: utf-8 -*-
u"""
A diagnostic utility that can be used to render email messages to files on disk.
"""
from __future__ import absolute_import, division, print_function

import errno
import logging
import os
from datetime import datetime

import attr
import six

from edx_ace.channel import Channel, ChannelType

LOG = logging.getLogger(__name__)

PATH_OVERRIDE_KEY = u'output_file_path'
DEFAULT_OUTPUT_FILE_PATH_TPL = u'/edx/src/ace_messages/{recipient.username}.{date:%Y%m%d-%H%M%S}.html'
TEMPLATE = u"""
<!DOCTYPE html>
<html>
    <head>
        {head_html}
    </head>
    <body>
        {body_html}
    </body>
    <!-- to: {email_address} -->
    <!-- from_name: {from_name} -->
    <!-- subject: {subject} -->
    <!-- body: {body} -->
</html>
"""
STDOUT_TEMPLATE = u"""
------- EMAIL -------
To: {email_address}
From: {from_name}
Subject: {subject}
Body:
{body}
-------  END  -------
"""
OUTPUT_ENCODING = u'utf8'


class FileEmailChannel(Channel):
    u"""
    An email channel that simply renders the message HTML to a file and the body text to stdout.

    If you add this channel to your enabled channels list as your email channel, it will write out the text version
    of the email to stdout and the HTML version to an output file.

    Examples::

        ACE_ENABLED_CHANNELS = ['file_email']

    By default it writes the output file to /edx/src/ace_output.html and overwrites any existing file at that location.
    In the edX devstack, this folder is shared between the host and the containers so you can easily open the file using
    a browser on the host. You can override this output file location by passing in a ``output_file_path`` key in the
    message options. That path specifies where in the container filesystem the file should be written.

    Both streams of output are UTF-8 encoded.
    """

    channel_type = ChannelType.EMAIL

    @classmethod
    def enabled(cls):
        u"""
        Returns: True always!
        """
        return True

    def _encode(self, s):  # pragma: no cover
        if six.PY2:
            return s.encode(OUTPUT_ENCODING)
        else:
            return s

    def deliver(self, message, rendered_message):
        template_vars = {k: v.strip() for k, v in attr.asdict(rendered_message).items()}
        template_vars[u'email_address'] = message.recipient.email_address

        rendered_template = TEMPLATE.format(**template_vars)
        output_file_path = message.options.get(PATH_OVERRIDE_KEY, DEFAULT_OUTPUT_FILE_PATH_TPL.format(
            recipient=message.recipient,
            date=datetime.now()
        ))
        make_parent_directories(output_file_path)
        with open(output_file_path, u'w') as output_file:
            output_file.write(self._encode(rendered_template))

        print(self._encode(STDOUT_TEMPLATE.format(**template_vars)))


def make_parent_directories(path):
    """ helper Method for creating parent directories. """
    parent_dir_path = os.path.dirname(os.path.realpath(path))
    try:
        os.makedirs(parent_dir_path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(parent_dir_path):
            pass
        else:
            raise



u"""
:mod:`edx_ace.policy` contains all classes relating to message policies.

These policies manage which messages should be sent over which channels,
and are a point of pluggability in ACE.
"""
from __future__ import absolute_import

from abc import ABCMeta, abstractmethod

import attr
import six

from django.conf import settings

from edx_ace.channel import ChannelType
from edx_ace.utils.once import once
from edx_ace.utils.plugins import get_plugins


@attr.s
class PolicyResult(object):
    u"""
    Arguments:
        deny (set): A set of :class:`.ChannelType` values that should be excluded
            when sending a message.
    """
    deny = attr.ib(
        default=attr.Factory(set),
    )

    @deny.validator
    def check_set_of_channel_types(self, attribute, set_value):
        for value in set_value:
            if value not in ChannelType:
                raise ValueError(u"PolicyResult for {} has an invalid value {}.".format(attribute, value))


POLICY_PLUGIN_NAMESPACE = u'openedx.ace.policy'


@six.add_metaclass(ABCMeta)
class Policy(object):
    u"""
    A ``Policy`` allows an application to specify what :class:`.Channel` any specific
    :class:`.Message` shouldn't be sent over. Policies are one of the primary
    extension mechanisms for ACE, and are registered using the entrypoint ``openedx.ace.policy``.
    """

    @classmethod
    def enabled(cls):
        return True

    @abstractmethod
    def check(self, message):
        u"""
        Validate the supplied :class:`~edx_ace.message.Message` against a specific
        delivery policy.

        Arguments:
            message (:class:`.Message`): The message to run the policy against.

        Returns: :class:`.PolicyResult`
            A :class:`.PolicyResult` that represents what channels the ``message`` should not be delivered over.
        """
        pass


# TODO(later): Need to support Denying ALL channels for a message
#   For example, a user (or course) policy denies sending all messages
#   of a particular type.

# TODO(later): We may also want to consider adding a separate
#   policy method for checking MessageType objects. For example,
#   the RecipientResolver could also check platform-wide (and
#   course-wide?) policies with a MessageType object.


def channels_for(message):
    u"""
    Arguments:
        message (:class:`.Message`): The message apply policies to.

    Returns: set
        A set of :class:`.ChannelType` values that are allowed by all policies
        applied to the message.
    """
    allowed_channels = set(ChannelType)

    for policy in policies():
        allowed_channels -= policy.check(message).deny

    message.report(u'policy_allowed_channels', u','.join(c.value for c in allowed_channels))
    return allowed_channels


@once
def policies():
    return [
        extension.obj
        for extension in get_plugins(
            namespace=POLICY_PLUGIN_NAMESPACE,
            names=getattr(settings, u'ACE_ENABLED_POLICIES', []),
        )
    ]



# -*- coding: utf-8 -*-
u"""
:mod:`edx_ace.utils.date` contains utility functions used for
serializing and deserializing dates. It is intended for internal ACE
use.
"""
from __future__ import absolute_import, division, print_function

from datetime import datetime

from dateutil.parser import parse
from dateutil.tz import tzutc


def get_current_time():
    u"""The current time in the UTC timezone as a timezone-aware datetime object."""
    return datetime.now(tzutc())


def serialize(timestamp_obj):
    u"""
    Serialize a datetime object to an ISO8601 formatted string.

    Args:
        timestamp_obj (datetime): The timestamp to serialize.

    Returns:
        basestring: A string representation of the timestamp in ISO8601 format.
    """
    if timestamp_obj is None:
        return None

    # TODO(later): my understanding is that type checks like this are not pythonic.
    assert isinstance(timestamp_obj, datetime)

    # TODO(later): what if utcoffset() returns 0?
    if timestamp_obj.tzinfo is not None and timestamp_obj.utcoffset() is None:
        return timestamp_obj.isoformat() + u'Z'
    else:
        return timestamp_obj.isoformat()


def deserialize(timestamp_iso8601_str):
    u"""
    Deserialize a datetime object from an ISO8601 formatted string.

    Args:
        timestamp_iso8601_str (basestring): A timestamp as an ISO8601 formatted string.

    Returns:
        datetime: A timezone-aware python datetime object.
    """
    if timestamp_iso8601_str is None:
        return None
    return parse(timestamp_iso8601_str, yearfirst=True)



# -*- coding: utf-8 -*-
u"""
:mod:`edx_ace.channel` exposes the ACE extension point needed
to add new delivery :class:`Channel` instances to an ACE application.

Developers wanting to add a new deliver channel should subclass :class:`Channel`,
and then add an entry to the ``openedx.ace.channel`` entrypoint in their ``setup.py``.
"""
from __future__ import absolute_import, division, print_function

import abc
from collections import OrderedDict, defaultdict
from enum import Enum

import six

from django.conf import settings

from edx_ace.errors import UnsupportedChannelError
from edx_ace.utils.once import once
from edx_ace.utils.plugins import get_plugins

# TODO(later): encapsulate the shared part of this namespace in the utils.plugin module
CHANNEL_EXTENSION_NAMESPACE = u'openedx.ace.channel'


@six.python_2_unicode_compatible
class ChannelType(Enum):
    u"""
    All supported communication channels.
    """
    EMAIL = u'email'
    PUSH = u'push'

    def __str__(self):
        return self.value


@six.add_metaclass(abc.ABCMeta)
class Channel(object):
    u"""
    Channels deliver messages to users that have already passed through the presentation and policy steps.

    Examples include email messages, push notifications, or in-browser messages. Implementations of this abstract class
    should not require any parameters be passed into their constructor since they are instantiated.

    :attr:`.channel_type` must be a :class:`.ChannelType`.
    """

    channel_type = None

    @classmethod
    def enabled(cls):
        u"""
        Validate settings to determine whether this channel can be enabled.
        """
        return True

    @abc.abstractmethod
    def deliver(self, message, rendered_message):
        u"""
        Transmit a rendered message to a recipient.

        Args:
            message (Message): The message to transmit.
            rendered_message (dict): The rendered content of the message that has been personalized for this particular
                recipient.
        """
        raise NotImplementedError()


@six.python_2_unicode_compatible
class ChannelMap(object):
    u"""
    A class that represents a channel map, usually as described in Django settings and `setup.py` files.
    """
    def __init__(self, channels_list):
        u"""
        Initialize a ChannelMap.

        Args:
            channels_list (list): A list of [channel_name, channel] pairs to fill in the channel map, in order.
        """
        self.channel_type_to_channel_impl = defaultdict(OrderedDict)
        for channel_name, channel in channels_list:
            self.register_channel(channel, channel_name)

    def register_channel(self, channel, channel_name):
        u"""
        Registers a channel in the channel map.

        Args:
            channel (Channel): The channel to register.
            channel_name (str): The channel name, as stated in the `setup.py` file.
        """
        self.channel_type_to_channel_impl[channel.channel_type][channel_name] = channel

    def get_channel_by_name(self, channel_type, channel_name):
        u"""
        Gets a registered a channel by its name and type.

        Raises:
            KeyError: If either of the channel or its type are not registered.

        Returns:
            Channel: The channel object.
        """
        return self.channel_type_to_channel_impl[channel_type][channel_name]

    def get_default_channel(self, channel_type):
        u"""
        Returns the first registered channel by type.

        Raises:
            UnsupportedChannelError: If there's no channel that matched the request.

        Args:
            channel_type (ChannelType): The channel type.
        """
        try:
            return six.next(six.itervalues(self.channel_type_to_channel_impl[channel_type]))
        except (StopIteration, KeyError):
            raise UnsupportedChannelError((
                u'No implementation for channel {channel_type} is registered. '
                u'Available channels are: {channels}'
            ).format(channel_type=channel_type, channels=channels()))

    def __str__(self):
        return u'ChannelMap {channels}'.format(channels=repr(self.channel_type_to_channel_impl))


@once
def channels():  # pragma: no cover
    u"""
    Gathers all available channels.

    Note that this function loads all available channels from entry points. It expects the Django setting
    ``ACE_ENABLED_CHANNELS`` to be a list of plugin names that should be enabled. Only one plugin per channel type
    should appear in that list.

    Raises:
        ValueError: If multiple plugins are enabled for the same channel type.

    Returns:
        ChannelMap: A mapping of channel types to instances of channel objects that can be used to deliver messages.
    """
    plugins = get_plugins(
        namespace=CHANNEL_EXTENSION_NAMESPACE,
        names=getattr(settings, u'ACE_ENABLED_CHANNELS', []),
    )

    return ChannelMap([
        [extension.name, extension.obj]
        for extension in plugins
    ])


def get_channel_for_message(channel_type, message):
    u"""
    Based on available `channels()` returns a single channels for a message.

    Raises:
        UnsupportedChannelError: If there's no channel matches the request.

    Returns:
        Channel: The selected channel object.
    """
    channels_map = channels()

    if channel_type == ChannelType.EMAIL:
        if message.options.get(u'transactional'):
            channel_name = settings.ACE_CHANNEL_TRANSACTIONAL_EMAIL
        else:
            channel_name = settings.ACE_CHANNEL_DEFAULT_EMAIL
        try:
            return channels_map.get_channel_by_name(channel_type, channel_name)
        except KeyError:
            pass

    return channels_map.get_default_channel(channel_type)



u"""
:mod:`edx_ace.recipient` contains :class:`Recipient`, which captures all targeting
information needed to deliver a message to some user.
"""
from __future__ import absolute_import

import attr

from edx_ace.serialization import MessageAttributeSerializationMixin


@attr.s
class Recipient(MessageAttributeSerializationMixin):
    u"""
    The target for a message.

    Arguments:
        username (str): The username of the intended recipient.
        email_address (str): The email address of the intended recipient. Optional.
    """
    username = attr.ib()
    email_address = attr.ib(default=None)



# -*- coding: utf-8 -*-
u"""
An internal module that manages the presentation/rendering step of the
ACE pipeline.
"""
from __future__ import absolute_import

from django.utils import translation

from edx_ace import errors, renderers
from edx_ace.channel import ChannelType

RENDERERS = {
    ChannelType.EMAIL: renderers.EmailRenderer(),
}


def render(channel, message):
    u""" Returns the rendered content for the given channel and message. """
    renderer = RENDERERS.get(channel.channel_type)

    if not renderer:
        error_msg = u'No renderer is registered for the channel type [{}].'.format(channel.channel_type)
        raise errors.UnsupportedChannelError(error_msg)

    message_language = message.language or translation.get_language()
    with translation.override(message_language):
        return renderer.render(channel, message)



# -*- coding: utf-8 -*-
u"""
:mod:`edx_ace.channel.django_email` implements a Django `send_mail()` email
delivery channel for ACE.
"""
from __future__ import absolute_import, division, print_function

import logging
import re
from smtplib import SMTPException

from django.conf import settings
from django.core.mail import EmailMultiAlternatives

from edx_ace.channel import Channel, ChannelType
from edx_ace.errors import FatalChannelDeliveryError

LOG = logging.getLogger(__name__)

TEMPLATE = u"""\
<!DOCTYPE html>
<html>
    <head>
        {head_html}
    </head>
    <body>
        {body_html}
    </body>
</html>
"""


class DjangoEmailChannel(Channel):
    u"""
    A `send_mail()` channel for edX ACE.

    This is both useful for providing an alternative to Sailthru and to debug ACE mail by
    inspecting `django.core.mail.outbox`.



    Example:

        Sample settings::

            .. settings_start
            EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
            EMAIL_HOST = 'localhost'
            DEFAULT_FROM_EMAIL = 'hello@example.org'

            ACE_CHANNEL_DEFAULT_EMAIL = 'sailthru_email'
            ACE_CHANNEL_TRANSACTIONAL_EMAIL = 'django_email'

            ACE_ENABLED_CHANNELS = [
                'sailthru_email',
                'django_email',
            ]
            .. settings_end
    """

    channel_type = ChannelType.EMAIL

    @classmethod
    def enabled(cls):
        u"""
        Returns: True always!
        """
        return True

    def deliver(self, message, rendered_message):
        # Compress spaces and remove newlines to make it easier to author templates.
        subject = re.sub(u'\\s+', u' ', rendered_message.subject, re.UNICODE).strip()
        default_from_address = getattr(settings, u'DEFAULT_FROM_EMAIL', None)
        reply_to = message.options.get(u'reply_to', None)
        from_address = message.options.get(u'from_address', default_from_address)
        if not from_address:
            raise FatalChannelDeliveryError(
                u'from_address must be included in message delivery options or as the DEFAULT_FROM_EMAIL settings'
            )

        rendered_template = TEMPLATE.format(
            head_html=rendered_message.head_html,
            body_html=rendered_message.body_html,
        )
        try:
            mail = EmailMultiAlternatives(
                subject=subject,
                body=rendered_message.body,
                from_email=from_address,
                to=[message.recipient.email_address],
                reply_to=reply_to,
            )

            mail.attach_alternative(rendered_template, u'text/html')
            mail.send()
        except SMTPException as e:
            LOG.exception(e)
            raise FatalChannelDeliveryError(u'An SMTP error occurred (and logged) from Django send_email()')
